"""Provide a model for Z-Wave controller firmware."""
from enum import IntEnum
from typing import TypedDict


class ControllerFirmwareUpdateStatus(IntEnum):
    """Enum with all controller firmware update status values.

    https://zwave-js.github.io/node-zwave-js/#/api/node?id=status
    """

    ERROR_TIMEOUT = 0
    # The maximum number of retry attempts for a firmware fragments were reached
    ERROR_RETRY_LIMIT_REACHED = 1
    # The update was aborted by the bootloader
    ERROR_ABORTED = 2
    OK = 255


class ControllerFirmwareUpdateProgressDataType(TypedDict):
    """Represent a controller firmware update progress dict type."""

    sentFragments: int
    totalFragments: int
    progress: float


class ControllerFirmwareUpdateProgress:
    """Model for a controller firmware update progress data."""

    def __init__(self, data: ControllerFirmwareUpdateProgressDataType) -> None:
        """Initialize."""
        self.data = data

    @property
    def sent_fragments(self) -> int:
        """Return the number of fragments sent to the device so far."""
        return self.data["sentFragments"]

    @property
    def total_fragments(self) -> int:
        """Return the total number of fragments that need to be sent to the device."""
        return self.data["totalFragments"]

    @property
    def progress(self) -> float:
        """Return progress."""
        return float(self.data["progress"])


class ControllerFirmwareUpdateResultDataType(TypedDict):
    """Represent a controller firmware update result dict type."""

    status: int
    success: bool


class ControllerFirmwareUpdateResult:
    """Model for controller firmware update result data."""

    def __init__(self, data: ControllerFirmwareUpdateResultDataType) -> None:
        """Initialize."""
        self.data = data

    @property
    def status(self) -> ControllerFirmwareUpdateStatus:
        """Return the firmware update status."""
        return ControllerFirmwareUpdateStatus(self.data["status"])

    @property
    def success(self) -> bool:
        """Return whether the firmware update was successful."""
        return self.data["success"]



"""Firmware update helper."""
import asyncio
from typing import Dict, List, Optional, cast

import aiohttp

from .client import Client
from .model.firmware import FirmwareUpdateData
from .model.node import Node


async def update_firmware(
    url: str,
    node: Node,
    updates: List[FirmwareUpdateData],
    session: aiohttp.ClientSession,
    additional_user_agent_components: Optional[Dict[str, str]] = None,
) -> bool:
    """Send updateFirmware command to Node."""
    client = Client(
        url, session, additional_user_agent_components=additional_user_agent_components
    )
    await client.connect()
    await client.initialize()

    receive_task = asyncio.get_running_loop().create_task(client.receive_until_closed())

    cmd = {
        "command": "node.update_firmware",
        "nodeId": node.node_id,
        "updates": [update.to_dict() for update in updates],
    }

    data = await client.async_send_command(cmd, require_schema=24)
    await client.disconnect()
    if not receive_task.done():
        receive_task.cancel()

    return cast(bool, data["success"])


async def controller_firmware_update_otw(
    url: str,
    firmware_file: FirmwareUpdateData,
    session: aiohttp.ClientSession,
    additional_user_agent_components: Optional[Dict[str, str]] = None,
) -> bool:
    """
    Send firmwareUpdateOTW command to Controller.

    Sending the wrong firmware to a controller can brick it and make it unrecoverable.
    Consumers of this library should build mechanisms to ensure that users understand
    the risks.
    """
    client = Client(
        url, session, additional_user_agent_components=additional_user_agent_components
    )
    await client.connect()
    await client.initialize()

    receive_task = asyncio.get_running_loop().create_task(client.receive_until_closed())

    data = await client.async_send_command(
        {
            "command": "controller.firmware_update_otw",
            **firmware_file.to_dict(),
        },
        require_schema=25,
    )
    await client.disconnect()
    if not receive_task.done():
        receive_task.cancel()

    return cast(bool, data["success"])



"""Provide a model for Z-Wave firmware."""
from dataclasses import dataclass
from typing import Optional

from ..const import TYPING_EXTENSION_FOR_TYPEDDICT_REQUIRED
from ..util.helpers import convert_bytes_to_base64

if TYPING_EXTENSION_FOR_TYPEDDICT_REQUIRED:
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class FirmwareUpdateDataDataType(TypedDict, total=False):
    """Represent a firmware update data dict type."""

    filename: str  # required
    file: str  # required
    fileFormat: str


@dataclass
class FirmwareUpdateData:
    """Firmware update data."""

    filename: str
    file: bytes
    file_format: Optional[str] = None

    def to_dict(self) -> FirmwareUpdateDataDataType:
        """Convert firmware update data to dict."""
        data: FirmwareUpdateDataDataType = {
            "filename": self.filename,
            "file": convert_bytes_to_base64(self.file),
        }
        if self.file_format is not None:
            data["fileFormat"] = self.file_format
        return data



"""Test the firmware update helper."""
from unittest.mock import patch

from zwave_js_server.firmware import controller_firmware_update_otw, update_firmware
from zwave_js_server.model.firmware import FirmwareUpdateData


async def test_update_firmware_guess_format(url, client_session, multisensor_6):
    """Test update_firmware with guessed format."""
    with patch("zwave_js_server.firmware.Client.connect") as connect_mock, patch(
        "zwave_js_server.firmware.Client.initialize"
    ) as initialize_mock, patch(
        "zwave_js_server.firmware.Client.async_send_command"
    ) as cmd_mock, patch(
        "zwave_js_server.firmware.Client.disconnect"
    ) as disconnect_mock:
        node = multisensor_6
        cmd_mock.return_value = {"success": True}
        assert await update_firmware(
            url, node, [FirmwareUpdateData("test", bytes(10))], client_session
        )

        connect_mock.assert_called_once()
        initialize_mock.assert_called_once()
        cmd_mock.assert_called_once_with(
            {
                "command": "node.update_firmware",
                "nodeId": node.node_id,
                "updates": [
                    {
                        "filename": "test",
                        "file": "AAAAAAAAAAAAAA==",
                    }
                ],
            },
            require_schema=24,
        )
        disconnect_mock.assert_called_once()


async def test_update_firmware_known_format_and_target(
    url, client_session, multisensor_6
):
    """Test update_firmware with known format."""
    with patch("zwave_js_server.firmware.Client.connect") as connect_mock, patch(
        "zwave_js_server.firmware.Client.initialize"
    ) as initialize_mock, patch(
        "zwave_js_server.firmware.Client.async_send_command"
    ) as cmd_mock, patch(
        "zwave_js_server.firmware.Client.disconnect"
    ) as disconnect_mock:
        node = multisensor_6
        cmd_mock.return_value = {"success": True}
        assert await update_firmware(
            url=url,
            node=node,
            updates=[FirmwareUpdateData("test", bytes(10), "test")],
            session=client_session,
        )

        connect_mock.assert_called_once()
        initialize_mock.assert_called_once()
        cmd_mock.assert_called_once_with(
            {
                "command": "node.update_firmware",
                "nodeId": node.node_id,
                "updates": [
                    {
                        "filename": "test",
                        "file": "AAAAAAAAAAAAAA==",
                        "fileFormat": "test",
                    }
                ],
            },
            require_schema=24,
        )
        disconnect_mock.assert_called_once()


async def test_controller_firmware_update_otw_guess_format(url, client_session):
    """Test controller_firmware_update_otw with guessed format."""
    with patch("zwave_js_server.firmware.Client.connect") as connect_mock, patch(
        "zwave_js_server.firmware.Client.initialize"
    ) as initialize_mock, patch(
        "zwave_js_server.firmware.Client.async_send_command"
    ) as cmd_mock, patch(
        "zwave_js_server.firmware.Client.disconnect"
    ) as disconnect_mock:
        cmd_mock.return_value = {"success": True}
        assert await controller_firmware_update_otw(
            url, FirmwareUpdateData("test", bytes(10)), client_session
        )

        connect_mock.assert_called_once()
        initialize_mock.assert_called_once()
        cmd_mock.assert_called_once_with(
            {
                "command": "controller.firmware_update_otw",
                "filename": "test",
                "file": "AAAAAAAAAAAAAA==",
            },
            require_schema=25,
        )
        disconnect_mock.assert_called_once()


async def test_controller_firmware_update_otw_known_format_and_target(
    url, client_session
):
    """Test controller_firmware_update_otw with known format."""
    with patch("zwave_js_server.firmware.Client.connect") as connect_mock, patch(
        "zwave_js_server.firmware.Client.initialize"
    ) as initialize_mock, patch(
        "zwave_js_server.firmware.Client.async_send_command"
    ) as cmd_mock, patch(
        "zwave_js_server.firmware.Client.disconnect"
    ) as disconnect_mock:
        cmd_mock.return_value = {"success": True}
        assert await controller_firmware_update_otw(
            url=url,
            firmware_file=FirmwareUpdateData("test", bytes(10), "test"),
            session=client_session,
        )

        connect_mock.assert_called_once()
        initialize_mock.assert_called_once()
        cmd_mock.assert_called_once_with(
            {
                "command": "controller.firmware_update_otw",
                "filename": "test",
                "file": "AAAAAAAAAAAAAA==",
                "fileFormat": "test",
            },
            require_schema=25,
        )
        disconnect_mock.assert_called_once()
